/*  Pasang Emas. Enjoy a unique traditional game of Brunei.
    Copyright (C) 2010  Nor Jaidi Tuah

    This file is part of Pasang Emas.
      
    Pasang Emas is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
namespace Pasang {

class Client : Object {
    private Gtk.Dialog dialog = null;

    /**
     * Signals relevant for game state.
     */
    public signal void on_join_game ();
    public signal void on_leave_game ();
    public signal void on_move_received (int move_sequence_num, string move_notation);

    /**
     * List of players logged in
     */   
    private Gtk.TreeView tree_view;
    private Gtk.TreeSelection tree_selection;
    private Gtk.ListStore player_list;

    /**
     * User entries
     */
    private Gtk.Entry server_address_entry = new Gtk.Entry ();
    private Gtk.Entry server_port_entry = new Gtk.Entry ();
    private Gtk.Entry user_name_entry = new Gtk.Entry ();

    /**
     * Command buttons
     */
    Gtk.Button logout_button = new Gtk.Button.with_label (_("Log Out"));
    Gtk.Button login_button = new Gtk.Button.with_label (_("Log In"));
    Gtk.Button challenge_button = new Gtk.Button.with_label (_("Challenge"));
    Gtk.Button retreat_button = new Gtk.Button.with_label (_("Retreat"));
     
    /**
     * Feedback
     */
    private Gtk.InfoBar status_info_bar = new Gtk.InfoBar ();
    private Gtk.Label status_info = new Gtk.Label (null);
    private string? error_message = null;

    /**
     * Communication with the asynchronous method 'dispatcher'
     */
    private Cancellable dispatcher_cancellable = new Cancellable ();
    private Queue<string> dispatcher_queue = new Queue<string> ();
    private SocketConnection connection = null;
    private DataInputStream input = null;
    private DataOutputStream output = null;

    /**
     * Information confirmed from the server
     */
    private long id = -1;
    private string user_name = null;
    private long opponent_id = -1;
    private string opponent_name = null;

    public Client () {
        dispatcher.begin ();
    }

    public void show_dialog () {
        create_dialog ();
        dialog.show ();
    }

    private void send (string message) {
        if (connection != null) {
            try {
                output.put_string (message, null);
            }
            catch (Error err) {
                stderr.printf ("Send error: %s\n", err.message);
            }
        }
    }

    public void send_move (int seq, string move_notation) {
        if (opponent_id == -1) return;
        send ("MOVE %d %s\n".printf (seq, move_notation));
    }

    private void dispatch (string command) {
        dispatcher_queue.push_tail (command);
        dispatcher_cancellable.cancel ();
    }

    private async void dispatcher () {
        while (true) {
            dispatcher_cancellable.reset ();
            if (dispatcher_queue.is_empty () && connection == null) {
                // Previously we used Idle.add.
                // The resulting busy waiting causes significant cpu
                // consumption (12% on my machine. Now, only 0 - 2%)
                yield Util.wait_async (1000);
                continue;
            }
            switch (dispatcher_queue.pop_head ()) {
                case "login" : yield login (); break;
                case "logout" : yield logout (); break;
                default : yield process_messages (); break;
            }
        }
    }

    /**
     * Attempt login.
     * input: server_address_entry, server_port_entry, user_name_entry
     * output: connection (null if fail), input, output
     * messages sent: LOGIN ...
     */
    private async void login () {
        try {
            dialog.get_window().cursor = new Gdk.Cursor.for_display (
                Gdk.Display.get_default (), Gdk.CursorType.WATCH);
            if (connection != null) yield logout ();
            var server_name = server_address_entry.text;
            var server_port = int.parse (server_port_entry.text);
            var resolver = Resolver.get_default ();
            var addresses = yield resolver.lookup_by_name_async (server_name, dispatcher_cancellable);
            var address = addresses.nth_data (0);
            debug ("server IP: %s", address.to_string ());
            var socket_address = new InetSocketAddress (address, (uint16) server_port);
            var client = new SocketClient ();
            connection = yield client.connect_async (socket_address, dispatcher_cancellable);
            input = new DataInputStream (connection.input_stream);
            output = new DataOutputStream (connection.output_stream);
            send ("LOGIN %s\n".printf (user_name_entry.text));
        } catch (IOError.CANCELLED err_cancelled) {
            connection = null;
        } catch (Error err) {
            connection = null;
            error_message = err.message;
        }
        update_status ();
        dialog.get_window().cursor = null;
    }

    /**
     * Destroy connection.
     * output: set connection to null
     */
    public async void logout () {
        if (dialog == null) return;
        try {
            if (connection != null) {
                yield connection.close_async (100, dispatcher_cancellable);
                connection = null;
            }
        }
        catch (Error err) {
            stderr.printf ("Logout error: %s\n", err.message);
        }
        player_list.clear ();
        user_name = null;
        id = -1;
        opponent_id = -1;
        on_leave_game ();
        update_status ();
    }

    /**
     * Issue a challenge.
     * input: list selection
     * message sent: CHALLENGE id
     */
    private void challenge () {
        if (tree_selection.count_selected_rows () != 1) return;
        Gtk.TreeIter iter;
        Gtk.TreeModel model;
        tree_selection.get_selected (out model, out iter);
        Value val;
        player_list.get_value (iter, 0, out val);
        send ("CHALLENGE %ld\n".printf (val.get_long ()));
    }

    /**
     * Process messages from the server.
     * The read loop may break if
     *  1. dispatcher_cancellable is triggered, or
     *  2. the connection is lost.
     */
    private async void process_messages () {
        if (connection == null) return;
        try {
            string line;
            size_t length;
            while ((line = yield input.read_line_async (100, dispatcher_cancellable, out length)) != null) {
                message ("From server: %s", line);
                interpret (sanitize (line));
                if (dispatcher_cancellable.is_cancelled ()) break;
            }
            // Disconnected from the server? If yes, logout.
            if (!dispatcher_cancellable.is_cancelled ()) {
                error_message = _("Disconnected from the server.");
                update_status ();
                yield logout ();
            }
        }
        catch (IOError.CANCELLED err_cancelled) {
            // A user command has been dispatched and must be attended
        }
        catch (Error err) {
            stderr.printf ("IO error: %s\n", err.message);
        }
    }

    private string sanitize (string s) {
        return s.strip().replace ("\n", "");
    }

    private void interpret (string line) throws Error {
        if (line == "") return;
        var msg = line.split (" ", 2);
        var arg = msg.length == 2 ? msg[1].strip () : "";
        switch (msg[0]) {
            case "LOGIN-OK" : login_accepted (arg); break;
            case "LOGIN-ERROR" : login_refused (); break;
            case "USER" : remote_user_registered (arg); break;
            case "EXIT" : remote_user_deregistered (arg); break;
            case "PLAY-AGAINST" : opponent_booked (arg); break;
            case "PLAY-END" : play_ended (); break;
            case "MOVE" : move_received (arg); break;
        }
    }

    /**
     * input: login_info in the form of "id name"
     * output: id, user_name
     * message sent: GET-USER-LIST
     * 
     */
    private void login_accepted (string login_info) throws Error {
        error_message = null;
        var info = login_info.split (" ", 2);
        if (info.length != 2) return;   // Malformed message from the server?
        id = long.parse (info[0]);
        user_name = info[1].strip ();
        send ("GET-USER-LIST\n");
        update_status ();
    }

    private void login_refused () {
        error_message = _("Login refused.");
        update_status ();
    }

    /**
     * Store a row of information inside player_list at the position given by iter.
     */
    private void set_row (Gtk.TreeIter iter, long row_id, char row_status, string row_name) {
        player_list.@set (iter, 
            0, row_id,
            1, row_status,
            2, row_name,
            3, row_status == 'p' ? _("Playing") : _("Waiting"),
            -1);
    }

    /**
     * Update list of users whenever a remote player logs in or changes her status.
     * Input: user_info in the form of "id status name"
     * Output: modified row, if id found; a new row if id is not found
     */
    private void remote_user_registered (string user_info) {
        var info = user_info.split (" ", 3);
        if (info.length != 3) return;
        var remote_user_id = long.parse (info[0]);
        if (remote_user_id == id) return; // Ignore if this is our own id
        var remote_user_status = info[1] == "p" ? 'p' : 'w';
        var remote_user_name = info[2].strip ();
        Gtk.TreeIter iter;
        var has_next = player_list.get_iter_first (out iter);
        while (has_next) {
            Value val;
            player_list.get_value (iter, 0, out val);
            if (val.get_long () == remote_user_id) {
                set_row (iter, remote_user_id, remote_user_status, remote_user_name);
                update_status ();
                return;
            }
            has_next = player_list.iter_next (ref iter);
        }
        // New player
        player_list.append (out iter);
        set_row (iter, remote_user_id, remote_user_status, remote_user_name);
        update_status ();
    }

    /**
     * Remove remote user entry.
     * Input: remote user id
     * Output: row deleted if found
     */
    private void remote_user_deregistered (string id_string) {
        var remote_user_id = long.parse (id_string);
        Gtk.TreeIter iter;
        var has_next = player_list.get_iter_first (out iter);
        while (has_next) {
            Value val;
            player_list.get_value (iter, 0, out val);
            if (val.get_long () == remote_user_id) {
                player_list.remove (iter);
                break;
            }
            has_next = player_list.iter_next (ref iter);
        }
        if (remote_user_id == opponent_id) {
            play_ended ();
        }
        update_status ();
    }

    /**
     * Receive an opponent.
     * Input: remote user id
     * output: opponent_id, opponent_name
     */
    private void opponent_booked (string id_string) {
        var remote_user_id = long.parse (id_string);
        if (remote_user_id == id) return;  // Our own id. Ignore.
        // Get opponent's name
        Gtk.TreeIter iter;
        var has_next = player_list.get_iter_first (out iter);
        while (has_next) {
            Value val;
            player_list.get_value (iter, 0, out val);
            if (val.get_long () == remote_user_id) {
                opponent_id = remote_user_id;
                on_join_game ();
                player_list.get_value (iter, 2, out val);
                opponent_name = val.get_string ();
                update_status ();
                break;
            }
            has_next = player_list.iter_next (ref iter);
        }
    }

    /**
     * Play ended. Retreat from opponnet.
     */
    private void play_ended () {
        opponent_id = -1;
        on_leave_game ();
        update_status ();
    }

    /**
     * Opponent's move has arrived.
     * input: opponent id + sequence number + move notation
     * output: signal emitted
     */
    private void move_received (string arg) {
        if (opponent_id == -1) {
            message ("Sync error: no opponent, but a move was received.");
            return;
        }
        var info = arg.split (" ", 3);
        if (info.length != 3) {
            message ("Format error: \"id seq move\" expected.");
            return;
        }
        if (opponent_id != long.parse (info[0])) {
            message ("Sync error: unexpected opponent id.");
            return;
        }
        on_move_received (int.parse (info[1]), info[2]);
    }

    
    /**
     * Update status visually.
     */
    private void update_status () {
        // Pass 1: Update button sensitivities
        // if not logged in yet:
        if (connection == null || user_name == null || id == -1) {
            login_button.sensitive = true;
            logout_button.sensitive = false;
            challenge_button.sensitive = false;
            retreat_button.sensitive = false;
        }
        // if already logged in, but not playing yet:
        else if (opponent_id == -1) {
            login_button.sensitive = false;
            logout_button.sensitive = true;
            retreat_button.sensitive = false;
            challenge_button.sensitive = false;
            // Check the chosen row. If the status is "waiting", enable the challenge command.
            if (tree_selection.count_selected_rows () == 1) {
                Gtk.TreeIter iter;
                Gtk.TreeModel model;
                tree_selection.get_selected (out model, out iter);
                Value val;
                player_list.get_value (iter, 1, out val);
                challenge_button.sensitive = val.get_schar () == 'w';
            }
        }
        // if currently playing:
        else {
            login_button.sensitive = false;
            logout_button.sensitive = true;
            challenge_button.sensitive = false;
            retreat_button.sensitive = true;
        }

        // Pass 2: Update info bar
        // if network error:
        if (error_message != null) {
            status_info_bar.message_type = Gtk.MessageType.ERROR;
            status_info.label = error_message;
            status_info_bar.show ();
        }
        // if not logged in yet:
        else if (connection == null || user_name == null || id == -1) {
            status_info_bar.hide ();
        }
        // if already logged in, but not playing yet:
        else if (opponent_id == -1) {
            if (player_list.iter_n_children (null) == 0) {
                status_info_bar.message_type = Gtk.MessageType.INFO;
                status_info.label = _("No other players are online");
                status_info_bar.show ();
            }
            else {
                status_info_bar.hide ();
            }
        }
        // if currently playing:
        else {
            status_info_bar.message_type = Gtk.MessageType.INFO;
            status_info.label = _("You are playing against %s.").printf (opponent_name);
            status_info_bar.show ();
        }
    }

    private void create_dialog () {
        if (dialog == null) {
            dialog = new Gtk.Dialog.with_buttons (
                _("Play Against Remote Player"),
                (GLib.Application.get_default () as Gtk.Application).active_window,
                Gtk.DialogFlags.USE_HEADER_BAR);
            dialog.delete_event.connect (() => {return dialog.hide_on_delete ();});

            // Server address and port
            var grid = new Gtk.Grid ();
            dialog.get_content_area () .add (grid);
            grid.set_row_spacing (5);
            grid.set_column_spacing (5);
            server_address_entry.text = "localhost";
            server_address_entry.width_chars = 25;
            server_port_entry.text = "3812";
            server_port_entry.width_chars = 6;
            grid.attach (new Gtk.Label (_("Server:")), 0, 0, 1, 1);
            grid.attach (server_address_entry, 1, 0, 1, 1);
            grid.attach (new Gtk.Label (_("Port:")), 2, 0, 1, 1);
            grid.attach (server_port_entry, 3, 0, 1, 1);
            server_address_entry.hexpand = true;

            // User name
            user_name_entry.text = "";
            user_name_entry.placeholder_text = _("Name seen by other players");
            user_name_entry.width_chars = 30;
            grid.attach (new Gtk.Label (_("Name:")), 0, 1, 1, 1);
            grid.attach (user_name_entry, 1, 1, 3, 1);

            // Login / Logout button
            var login_button_box = new Gtk.ButtonBox (Gtk.Orientation.HORIZONTAL);
            login_button_box.layout_style = Gtk.ButtonBoxStyle.END;
            grid.attach (login_button_box, 0, 2, 4, 1);
            login_button_box.add (logout_button);
            logout_button.clicked.connect (() => {dispatch ("logout");});
            login_button_box.add (login_button);
            login_button.clicked.connect (() => {dispatch ("login");});

            // Status
            status_info_bar.get_content_area () .add (status_info);
            status_info.wrap = true;
            grid.attach (status_info_bar, 0, 3, 4, 1);

            // List of players
            var list = create_player_list ();
            grid.attach (list, 0, 4, 4, 1);
            list.set_size_request (150, 230);
            list.vexpand = true;

            // Challenge button
            var war_button_box = new Gtk.ButtonBox (Gtk.Orientation.HORIZONTAL);
            war_button_box.layout_style = Gtk.ButtonBoxStyle.END;
            grid.attach (war_button_box, 0, 5, 4, 1);
            war_button_box.add (challenge_button);
            challenge_button.clicked.connect (() => {challenge ();});
            war_button_box.add (retreat_button);
            retreat_button.clicked.connect (() => {send ("LEAVE\n");});

            dialog.show_all ();
            dialog.hide ();
            update_status ();
        }
    }

    private Gtk.Widget create_player_list () {
        // Construct model and view for player list
        tree_view  = new Gtk.TreeView ();
        tree_view.model = player_list = new Gtk.ListStore (4,
            typeof (long),      // connection id
            typeof (char),      // status (w for waiting, p for playing)
            typeof (string),    // player name
            typeof (string)     // status (localized)
        );
        tree_view.insert_column_with_attributes (0, "Name", new Gtk.CellRendererText (), text:2);
        tree_view.insert_column_with_attributes (1, "Status", new Gtk.CellRendererText (), text:3);

        tree_selection = tree_view.get_selection ();
        tree_selection.changed.connect (update_status);

        // Set up visual
        tree_view.columns_autosize ();
        tree_view.set_headers_visible (false);
        tree_view.set_cursor (new Gtk.TreePath.from_string ("0"), null, false);
        var scrolled = new Gtk.ScrolledWindow (null, null);
        scrolled.set_policy (Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS);
        scrolled.add (tree_view);
        return scrolled;
    }
}//class Client
}//namespace
// vim: tabstop=4: expandtab: textwidth=100: autoindent:
